var RatingScore = (function (exports) {
  'use strict';

  function _toConsumableArray(arr) {
    return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
  }

  function _arrayWithoutHoles(arr) {
    if (Array.isArray(arr)) {
      for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

      return arr2;
    }
  }

  function _iterableToArray(iter) {
    if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
  }

  function _nonIterableSpread() {
    throw new TypeError("Invalid attempt to spread non-iterable instance");
  }

  /**
   * Rounds a number to two decimal places.
   *
   * @private
   * @param {Number} num - A number to round.
   * @returns {Number} Rounded number.
   */
  var roundToTwo = function roundToTwo(num) {
    return Math.round(num * 100) / 100;
  };
  /**
   * Clamps a number to 0-1 range.
   *
   * @private
   * @param {Number} num - A number to clamp.
   * @returns {Number} Clamped number.
   */

  var clamp = function clamp(num) {
    return Math.min(Math.max(num, 0), 1);
  };
  /**
   * Returns a curried equivalent of the provided function.
   *
   * @private
   * @param {Function} fn - The function to curry.
   * @returns {Function} A new, curried function.
   */

  var curry = function curry(fn) {
    var len = fn.length;
    var args = [];
    return function next() {
      args = _toConsumableArray(args).concat(Array.prototype.slice.call(arguments));
      return args.length >= len ? fn.apply(this, args) : next;
    };
  };

  /* @flow */

  /* @flow */

  /**
   * We use `ε`, epsilon, as a stopping criterion when we want to iterate
   * until we're "close enough". Epsilon is a very small number: for
   * simple statistics, that number is **0.0001**
   *
   * This is used in calculations like the binomialDistribution, in which
   * the process of finding a value is [iterative](https://en.wikipedia.org/wiki/Iterative_method):
   * it progresses until it is close enough.
   *
   * Below is an example of using epsilon in [gradient descent](https://en.wikipedia.org/wiki/Gradient_descent),
   * where we're trying to find a local minimum of a function's derivative,
   * given by the `fDerivative` method.
   *
   * @example
   * // From calculation, we expect that the local minimum occurs at x=9/4
   * var x_old = 0;
   * // The algorithm starts at x=6
   * var x_new = 6;
   * var stepSize = 0.01;
   *
   * function fDerivative(x) {
   *   return 4 * Math.pow(x, 3) - 9 * Math.pow(x, 2);
   * }
   *
   * // The loop runs until the difference between the previous
   * // value and the current value is smaller than epsilon - a rough
   * // meaure of 'close enough'
   * while (Math.abs(x_new - x_old) > ss.epsilon) {
   *   x_old = x_new;
   *   x_new = x_old - stepSize * fDerivative(x_old);
   * }
   *
   * console.log('Local minimum occurs at', x_new);
   */
  var epsilon = 0.0001;
  var LOGSQRT2PI = Math.log(Math.sqrt(2 * Math.PI));

  /* @flow */

  var SQRT_2PI = Math.sqrt(2 * Math.PI);

  /* @flow */

  var SQRT_2PI$1 = Math.sqrt(2 * Math.PI);

  /* @flow */

  /**
   * The Inverse [Gaussian error function](http://en.wikipedia.org/wiki/Error_function)
   * returns a numerical approximation to the value that would have caused
   * `errorFunction()` to return x.
   *
   * @param {number} x value of error function
   * @returns {number} estimated inverted value
   */
  function inverseErrorFunction(x/*: number */)/*: number */ {
      var a = (8 * (Math.PI - 3)) / (3 * Math.PI * (4 - Math.PI));

      var inv = Math.sqrt(Math.sqrt(
          Math.pow(2 / (Math.PI * a) + Math.log(1 - x * x) / 2, 2) -
          Math.log(1 - x * x) / a) -
          (2 / (Math.PI * a) + Math.log(1 - x * x) / 2));

      if (x >= 0) {
          return inv;
      } else {
          return -inv;
      }
  }

  /* @flow */

  /**
   * The [Probit](http://en.wikipedia.org/wiki/Probit)
   * is the inverse of cumulativeStdNormalProbability(),
   * and is also known as the normal quantile function.
   *
   * It returns the number of standard deviations from the mean
   * where the p'th quantile of values can be found in a normal distribution.
   * So, for example, probit(0.5 + 0.6827/2) ≈ 1 because 68.27% of values are
   * normally found within 1 standard deviation above or below the mean.
   *
   * @param {number} p
   * @returns {number} probit
   */
  function probit(p /*: number */)/*: number */ {
      if (p === 0) {
          p = epsilon;
      } else if (p >= 1) {
          p = 1 - epsilon;
      }
      return Math.sqrt(2) * inverseErrorFunction(2 * p - 1);
  }

  /**
   * Returns the number of standard deviations above or below the mean
   * by which the value can be found in a normal distribution with
   * a given confidence.
   *
   * @private
   * @param {Number} [confidence=0.95] - Probability as fraction or percentage.
   * @returns {Number} Calculated z-score.
   */

  var zScore$1 = function zScore$$1() {
    var num = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0.95;

    if (num === 0) {
      return 0;
    }

    var confidence = clamp(num > 1 ? num / 100 : num);
    return roundToTwo(probit(1 - (1 - confidence) / 2));
  };

  /**
   * Returns a curried score function with precalculated z-score.
   *
   * @private
   * @param {Function} fn - Function calculating score.
   * @returns {Function}
   */

  var createScoreFunction = function createScoreFunction(fn) {
    return function (confidence) {
      var zScore = zScore$1(confidence);

      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      return curry(fn).apply(void 0, [zScore].concat(args));
    };
  };

  /**
   * Calculates lower bound of Wilson score confidence interval for a Bernoulli parameter.
   * This is the score that we want to use to sort our things by.
   *
   * Read more in an article by Evan Miller
   * [How Not To Sort By Average Rating](@link https://www.evanmiller.org/how-not-to-sort-by-average-rating.html).
   *
   * @private
   * @param {Number} z - Z-score.
   * @param {Number} p - Number of positive reviews (upvotes).
   * @param {Number} n - Number of all reviews (upvotes + downvotes).
   * @returns {Number} Calculated score (lower bound).
   */
  var lowerBound$1 = function lowerBound(z, p, n) {
    var z2 = Math.pow(z, 2);
    var p̂ = p / n;
    return (p̂ + z2 / (2 * n) - z * Math.sqrt((p̂ * (1 - p̂) + z2 / (4 * n)) / n)) / (1 + z2 / n);
  };

  /**
   * Calculates the score from the average of all reviews.
   *
   * This works by converting the average to a number of positive
   * reviews, so the lower bound method can be used.
   *
   * @private
   * @param {Number} zScore
   * @param {Number} averageRating - Average review score.
   * @param {Number} maxRating - Maximum review score.
   * @param {Number} totalReviews - Total number of reviews.
   * @returns {Number} Calculated score (lower bound).
   */

  var fromAverage = function fromAverage(zScore, averageRating, maxRating, totalRatings) {
    return lowerBound$1(zScore, averageRating * totalRatings / maxRating, totalRatings);
  };

  /**
   * Sums reviews weighted by number of stars. The weight is calculated
   * by dividing number of stars (counting from zero) by the maximum
   * number of stars minus 1. In effect, the worst reviews get a weight
   * of 0, and full stars reviews get weight equal to 1.
   *
   * For example, for a 5 stars rating system the weights are as follows:
   * 5 stars - 1
   * 4 stars - 0.75
   * 3 stars - 0.50
   * 2 stars - 0.25
   * 1 star  - 0
   *
   * @private
   * @param {Number} result - Sum of weighted ratings.
   * @param {Number} reviews - Number of reviews with given number of stars.
   * @param {Number} stars - Number of stars (counts from 0).
   * @param {Number[]} starsList - List of review counts.
   * @returns {Number} Calculated score (lower bound).
   */

  var weightStars = function weightStars(result, reviews, stars, _ref) {
    var length = _ref.length;
    return result += reviews * stars / (length - 1);
  };
  /**
   * Calculates the score (lower bound) from a list of review counts.
   *
   * The lower bound method can't be used with a non-binary
   * rating systems, so the ratings are weighted and reduced
   * to a single number of "positive" reviews, where the neutral
   * reviews are counted as a fraction of a positive review.
   *
   * @private
   * @param {Number} zScore
   * @param {Number[]} starsList - List of review counts.
   * @returns {Number} Calculated score (lower bound).
   */


  var fromStars = function fromStars(zScore, starsList) {
    var positiveReviews = starsList.reduce(weightStars, 0);
    var totalReviews = starsList.reduce(function (a, b) {
      return a + b;
    }, 0);
    return lowerBound$1(zScore, positiveReviews, totalReviews);
  };

  /**
   * Calculates the score from the number of upvotes and downvotes.
   *
   * @private
   * @param {Number} zScore
   * @param {Number} upvotes - Number of positive reviews.
   * @param {Number} downvotes - Number of negative reviews.
   * @returns {Number} Calculated score (lower bound).
   */

  var fromUpvotes = function fromUpvotes(zScore, upvotes, downvotes) {
    return lowerBound$1(zScore, upvotes, upvotes + downvotes);
  };

  /**
   * Calculates the score from the average of all reviews.
   *
   * This function is curried.
   *
   * @function
   * @param {Number} confidence
   * @param {Number} averageRating - Average review score.
   * @param {Number} maxRating - Maximum review score.
   * @param {Number} totalReviews - Total number of reviews.
   * @returns {Number} Calculated score.
   */

  var averageToScore = createScoreFunction(fromAverage);
  /**
   * Calculates the score from a list of review counts.
   *
   * This function is curried.
   *
   * @function
   * @param {Number} confidence
   * @param {Number} starsList - List of review counts.
   * @returns {Number} Calculated score.
   */

  var starsToScore = createScoreFunction(fromStars);
  /**
   * Calculates the score from the number of upvotes and downvotes.
   *
   * This function is curried.
   *
   * @function
   * @param {Number} confidence
   * @param {Number} upvotes - Number of positive reviews.
   * @param {Number} downvotes - Number of negative reviews.
   * @returns {Number} Calculated score.
   */

  var upvotesToScore = createScoreFunction(fromUpvotes);

  exports.averageToScore = averageToScore;
  exports.starsToScore = starsToScore;
  exports.upvotesToScore = upvotesToScore;

  return exports;

}({}));
