import { probit } from 'simple-statistics';

/**
 * Rounds a number to two decimal places.
 *
 * @private
 * @param {Number} num - A number to round.
 * @returns {Number} Rounded number.
 */
const roundToTwo = (num) => Math.round(num * 100) / 100;

/**
 * Clamps a number to 0-1 range.
 *
 * @private
 * @param {Number} num - A number to clamp.
 * @returns {Number} Clamped number.
 */
const clamp = (num) => Math.min(Math.max(num, 0), 1);

/**
 * Returns a curried equivalent of the provided function.
 *
 * @private
 * @param {Function} fn - The function to curry.
 * @returns {Function} A new, curried function.
 */
const curry = (fn) => {

  const len = fn.length;
  let args = [];

  return function next() {

    args = [...args, ...arguments];
    return (args.length >= len) ? fn.apply(this, args) : next
  }
};

/**
 * Returns the number of standard deviations above or below the mean
 * by which the value can be found in a normal distribution with
 * a given confidence.
 *
 * @private
 * @param {Number} [confidence=0.95] - Probability as fraction or percentage.
 * @returns {Number} Calculated z-score.
 */
const zScore = (num = 0.95) => {

  if (num === 0) {
    return 0
  }

  const confidence = clamp(num > 1 ? num / 100 : num);

  return roundToTwo(probit(1 - (1 - confidence) / 2))
};

/**
 * Returns a curried score function with precalculated z-score.
 *
 * @private
 * @param {Function} fn - Function calculating score.
 * @returns {Function}
 */
const createScoreFunction = (fn) => (confidence, ...args) => {

  const zScore$$1 = zScore(confidence);
  return curry(fn)(zScore$$1, ...args)
};

/**
 * Calculates lower bound of Wilson score confidence interval for a Bernoulli parameter.
 * This is the score that we want to use to sort our things by.
 *
 * Read more in an article by Evan Miller
 * [How Not To Sort By Average Rating](@link https://www.evanmiller.org/how-not-to-sort-by-average-rating.html).
 *
 * @private
 * @param {Number} z - Z-score.
 * @param {Number} p - Number of positive reviews (upvotes).
 * @param {Number} n - Number of all reviews (upvotes + downvotes).
 * @returns {Number} Calculated score (lower bound).
 */
const lowerBound = (z, p, n) => {

  const z2 = Math.pow(z, 2);
  const p̂ = p / n;

  return (
    (
      p̂ + z2 / (2 * n) - z *
      Math.sqrt((p̂ * (1 - p̂) + z2 / (4 * n)) / n)
    ) / (1 + z2 / n)
  )
};

/**
 * Calculates the score from the average of all reviews.
 *
 * This works by converting the average to a number of positive
 * reviews, so the lower bound method can be used.
 *
 * @private
 * @param {Number} zScore
 * @param {Number} averageRating - Average review score.
 * @param {Number} maxRating - Maximum review score.
 * @param {Number} totalReviews - Total number of reviews.
 * @returns {Number} Calculated score (lower bound).
 */
const fromAverage = (zScore, averageRating, maxRating, totalRatings) =>
  lowerBound(zScore, averageRating * totalRatings / maxRating, totalRatings);

/**
 * Sums reviews weighted by number of stars. The weight is calculated
 * by dividing number of stars (counting from zero) by the maximum
 * number of stars minus 1. In effect, the worst reviews get a weight
 * of 0, and full stars reviews get weight equal to 1.
 *
 * For example, for a 5 stars rating system the weights are as follows:
 * 5 stars - 1
 * 4 stars - 0.75
 * 3 stars - 0.50
 * 2 stars - 0.25
 * 1 star  - 0
 *
 * @private
 * @param {Number} result - Sum of weighted ratings.
 * @param {Number} reviews - Number of reviews with given number of stars.
 * @param {Number} stars - Number of stars (counts from 0).
 * @param {Number[]} starsList - List of review counts.
 * @returns {Number} Calculated score (lower bound).
 */
const weightStars = (result, reviews, stars, { length }) =>
  result += reviews * stars / (length - 1);

/**
 * Calculates the score (lower bound) from a list of review counts.
 *
 * The lower bound method can't be used with a non-binary
 * rating systems, so the ratings are weighted and reduced
 * to a single number of "positive" reviews, where the neutral
 * reviews are counted as a fraction of a positive review.
 *
 * @private
 * @param {Number} zScore
 * @param {Number[]} starsList - List of review counts.
 * @returns {Number} Calculated score (lower bound).
 */
const fromStars = (zScore, starsList) => {

  const positiveReviews = starsList.reduce(weightStars, 0);
  const totalReviews = starsList.reduce((a, b) => a + b, 0);

  return lowerBound(zScore, positiveReviews, totalReviews)
};

/**
 * Calculates the score from the number of upvotes and downvotes.
 *
 * @private
 * @param {Number} zScore
 * @param {Number} upvotes - Number of positive reviews.
 * @param {Number} downvotes - Number of negative reviews.
 * @returns {Number} Calculated score (lower bound).
 */
const fromUpvotes = (zScore, upvotes, downvotes) =>
  lowerBound(zScore, upvotes, upvotes + downvotes);

/**
 * Calculates the score from the average of all reviews.
 *
 * This function is curried.
 *
 * @function
 * @param {Number} confidence
 * @param {Number} averageRating - Average review score.
 * @param {Number} maxRating - Maximum review score.
 * @param {Number} totalReviews - Total number of reviews.
 * @returns {Number} Calculated score.
 */
const averageToScore = createScoreFunction(fromAverage);

/**
 * Calculates the score from a list of review counts.
 *
 * This function is curried.
 *
 * @function
 * @param {Number} confidence
 * @param {Number} starsList - List of review counts.
 * @returns {Number} Calculated score.
 */
const starsToScore = createScoreFunction(fromStars);

/**
 * Calculates the score from the number of upvotes and downvotes.
 *
 * This function is curried.
 *
 * @function
 * @param {Number} confidence
 * @param {Number} upvotes - Number of positive reviews.
 * @param {Number} downvotes - Number of negative reviews.
 * @returns {Number} Calculated score.
 */
const upvotesToScore = createScoreFunction(fromUpvotes);

export { averageToScore, starsToScore, upvotesToScore };
//# sourceMappingURL=rating-score.mjs.map
